/*
 * Nomad Instrument Control Software
 *
 * Copyright 2011 Institut Laue-Langevin
 *
 * Licensed under the EUPL, Version 1.1 only (the "License");
 * You may not use this work except in compliance with the Licence.
 * You may obtain a copy of the Licence at:
 *
 * http://joinup.ec.europa.eu/software/page/eupl
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the Licence is distributed on an "AS IS" basis,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the Licence for the specific language governing permissions and
 * limitations under the Licence.
 */

#include <drivers/ioxos/cadaracheacq/CadaracheAcqDef.h>
#include <drivers/ioxos/cadaracheacq/CadaracheAcqDriver.h>
#include <drivers/ioxos/cadaracheacq/RealCadaracheAcqDriver.h>
#include "Utilities/Calculations.h"

using namespace acquisition;
using namespace driver;
using namespace std;

namespace cadaracheacq {

/*
 * Constructor
 */
RealCadaracheAcqDriver::RealCadaracheAcqDriver(CadaracheAcqDriver* owner) :
				CadaracheAcqState(owner) {
}

/*
 * Destructor
 */
RealCadaracheAcqDriver::~RealCadaracheAcqDriver() {
	/* Empty */
}

/*
 * init
 */
void RealCadaracheAcqDriver::init() {
	readInfos();
}

/*
 * writeParam
 */
void RealCadaracheAcqDriver::writeParam() {
	int32 nbrepaints = owner()->nbRepaint.setpoint();
	int32 nbslices = owner()->nbSlicesAcq();
	float64* slices = owner()->slicesAcq();
	cout << "nbslices=" << nbslices << endl;
	owner()->write(NB_CHANNELS_REGISTER, Vme::LONGWORD_WIDTH, 1, (void*) &nbslices);
	owner()->write(NB_CYCLES_REGISTER, Vme::LONGWORD_WIDTH, 1, (void*) &nbrepaints);
	unique_ptr<int32> values(new int32[nbslices]);
	int32* buf = values.get();
	for(int32 i = 0; i < nbslices; ++ i) {
		buf[i] = (int32) slices[i] * 100; // ns
	}
	owner()->write(A32_ADDRESS, Vme::A32_ACCESS, TIMES_MEMORY , Vme::LONGWORD_WIDTH, nbslices, (void*) buf);
}

/*
 * readInfos
 */
void RealCadaracheAcqDriver::readInfos() {
	uint32 version;
	owner()->read(VERSION_REGISTER, Vme::LONGWORD_WIDTH, 1, (void*) &version);
	ostringstream buf;
	buf << hex << version << dec;
	owner()->version.update(buf.str());
}

/*
 * start
 */
void RealCadaracheAcqDriver::start() {
	uint32 write_function = 0;
	owner()->write(START_REGISTER, Vme::LONGWORD_WIDTH, 1, (void*) &write_function);
}

/*
 * clear
 */
void RealCadaracheAcqDriver::clear() {
	uint32 write_function = 0;
	owner()->write(CLEAR_RESULTS_REGISTER, Vme::LONGWORD_WIDTH, 1, (void*) &write_function);
	owner()->write(CLEAR_ALL_REGISTER, Vme::LONGWORD_WIDTH, 1, (void*) &write_function);
	owner()->write(CLEAR_MEMORIES_REGISTER, Vme::LONGWORD_WIDTH, 1, (void*) &write_function);
}

/*
 * pause
 */
void RealCadaracheAcqDriver::pause() {
}

/*
 * resume
 */
void RealCadaracheAcqDriver::resume() {
}

/*
 * stop
 */
void RealCadaracheAcqDriver::stop() {
}

/*
 * readParam
 */
void RealCadaracheAcqDriver::readParam() {
}

/*
 * synchroniseRead
 */
void RealCadaracheAcqDriver::synchroniseRead() {
	read();
	owner()->m_SynchroniseReadTerminated = true;
}

/*
 * read
 */
void RealCadaracheAcqDriver::read() {
	float64 time;
	uint32 vh, vl;
	owner()->read(ACTUAL_TIME_REGISTER, Vme::LONGWORD_WIDTH, 1, (void*) &vh);
	owner()->read(ACTUAL_TIME_REGISTER + 4, Vme::LONGWORD_WIDTH, 1, (void*) &vl);
	time = ((float64) vl + (float64) vh * pow(2, 32)) / 100000000.;
	owner()->time.update(time);

	owner()->read(ACTUAL_CHANNEL_REGISTER, Vme::LONGWORD_WIDTH, 1, (void*) &vl);
	owner()->actualSlice.update(vl);

	float64 sum = 0;
	int32* data = owner()->data();
	int32 nbslices = owner()->nbSlicesAcq();

	owner()->read(A32_ADDRESS, Vme::A32_ACCESS, DATA_MEMORY, Vme::LONGWORD_WIDTH, nbslices, (void*) data);

	for(int32 i = 0;i < nbslices;++i) {
		sum += data[i];
	}
	owner()->sum.update(sum);
	owner()->data.sendEvent();
}

/*
 * readStatus
 */
void RealCadaracheAcqDriver::readStatus() {
	int32 status;
	owner()->read(STATUS_REGISTER, Vme::LONGWORD_WIDTH, 1, (void*) &status);

	if (owner()->startActivated() == true) {
		if (((status & GATE_EXT_STATUS) == 0) && ((status & FUNCTION_START_ACTIVE_STATUS) != 0)) {
			// wait
		}
		else if (status & RUNNING_STATUS) {
			owner()->calculateProgression();
		} else {
			owner()->startActivated = false;
			owner()->commandProgression = PROGRESSION_END_DEVICE_CONTAINER;
		}
	} else {
		if (owner()->m_SynchroniseReadActivated == true) {
			if (owner()->m_SynchroniseReadTerminated == true) {
				owner()->m_SynchroniseReadActivated = false;
				owner()->m_SynchroniseReadTerminated = false;
				owner()->commandProgression = PROGRESSION_END_DEVICE_CONTAINER;
			}
		} else {
			if ((status & CLEAR_MEMORY_DATA_ACTIVE_STATUS) || (status & CLEAR_MEMORY_TIME_ACTIVE_STATUS)) {
				// wait
			}
			else {
				owner()->commandProgression = PROGRESSION_END_DEVICE_CONTAINER;
			}
		}
	}
	owner()->status.update(status);
	owner()->computeGateStatus(status);
	owner()->computeAcquisitionStatus(status);

}

}
